use reqwest::header::HeaderValue;
use reqwest::RequestBuilder;
use reqwest::Response;
use std::time::{SystemTime, UNIX_EPOCH};
use tokio::time::{sleep, Duration};

pub struct GithubEndpointWatcher {
    username: String,
    token: String,
    url: String,
    http_client: reqwest::Client,
    etag: Option<HeaderValue>,
    last_modified_at: Option<HeaderValue>,
    next_poll_allowed: u64,
    ratelimit_remaining: Option<u64>,
    ratelimit_reset: Option<u64>,
    previous_poll_interval: Option<u64>,
}

impl GithubEndpointWatcher {
    pub fn new(
        username: String,
        token: String,
        url: String,
    ) -> Result<Self, Box<dyn std::error::Error>> {
        let http_client = reqwest::Client::new();
        Ok(GithubEndpointWatcher {
            username,
            token,
            url,
            http_client,
            etag: None,
            last_modified_at: None,
            next_poll_allowed: 0,
            ratelimit_remaining: None,
            ratelimit_reset: None,
            previous_poll_interval: None,
        })
    }

    pub async fn skip_results_before_now(&mut self) -> Result<(), Box<dyn std::error::Error>> {
        // This will incur an API call to get the headers for ETAG and
        // last modified at but it will throw away the results.

        // TODO: Can this be done without a request? We wouldn't get
        // an ETAG but we should be able to calculate our own date
        // string for last modified.
        let request = self.build_request()?;
        self.sleep_until_next_poll().await?;
        self.sleep_until_ratelimit().await?;
        let request_started_at = SystemTime::now().duration_since(UNIX_EPOCH)?.as_secs();
        info!("Hitting url {}", self.url);
        let response = request.send().await?;
        self.update_headers_from_response(&response, request_started_at)?;
        response.error_for_status()?;

        Ok(())
    }

    fn build_request(&mut self) -> Result<RequestBuilder, Box<dyn std::error::Error>> {
        let mut request = self
            .http_client
            .get(&self.url)
            .basic_auth(self.username.trim(), Some(self.token.trim()))
            .header("User-Agent", "github_watcher")
            .header("Accept", "application/vnd.github.v3+json");
        if let Some(etag) = &self.etag {
            let trimmed = etag.to_str()?.trim_matches('"');
            info!("Setting etag to {}", trimmed);
            request = request.header(reqwest::header::ETAG, trimmed);
        }
        if let Some(last_modified_at) = &self.last_modified_at {
            // let trimmed = etag.to_str()?.trim_matches('"');
            info!(
                "Setting If-Modified-Since to {}",
                last_modified_at.to_str()?
            );
            request = request.header(reqwest::header::IF_MODIFIED_SINCE, last_modified_at);
        }
        Ok(request)
    }

    fn update_headers_from_response(
        &mut self,
        response: &Response,
        request_started_at: u64,
    ) -> Result<(), Box<dyn std::error::Error>> {
        let headers = response.headers();
        self.etag = headers.get(reqwest::header::ETAG).map(|x| x.to_owned());
        let poll_interval = number_header(headers.get("x-poll-interval"))?;
        if let Some(interval) = poll_interval {
            self.previous_poll_interval = Some(interval);
        }
        // let ratelimit_limit = headers.get("x-ratelimit-limit").map(|x| x.to_owned());
        self.ratelimit_remaining = number_header(headers.get("x-ratelimit-remaining"))?;
        self.ratelimit_reset = number_header(headers.get("x-ratelimit-reset"))?;
        // let ratelimit_used = headers.get("x-ratelimit-used").map(|x| x.to_owned());
        // let ratelimit_resource = headers.get("x-ratelimit-resource").map(|x| x.to_owned());
        self.last_modified_at = headers
            .get(reqwest::header::LAST_MODIFIED)
            .map(|x| x.to_owned());

        let poll_interval_parsed: u64 = poll_interval.unwrap_or_else(|| {
            let fallback_interval = self.previous_poll_interval.unwrap_or(300);
            info!(
                "No poll interval returned, defaulting to {} second polling.",
                fallback_interval
            );
            fallback_interval
        });
        self.next_poll_allowed = request_started_at + poll_interval_parsed;
        Ok(())
    }

    pub async fn get_results(
        &mut self,
    ) -> Result<Option<serde_json::Value>, Box<dyn std::error::Error>> {
        let request = self.build_request()?;
        self.sleep_until_next_poll().await?;
        self.sleep_until_ratelimit().await?;
        let request_started_at = SystemTime::now().duration_since(UNIX_EPOCH)?.as_secs();
        info!("Hitting url {}", self.url);
        let response = request.send().await?;

        self.update_headers_from_response(&response, request_started_at)?;
        let response_status = response.status();

        if let reqwest::StatusCode::NOT_MODIFIED = response_status {
            return Ok(None);
        } else {
            let body = response.json::<serde_json::Value>().await?;
            return Ok(Some(body));
        }
    }

    async fn sleep_until_next_poll(&self) -> Result<(), Box<dyn std::error::Error>> {
        loop {
            let now = SystemTime::now().duration_since(UNIX_EPOCH)?.as_secs();
            if now >= self.next_poll_allowed {
                break;
            }
            sleep(Duration::from_secs(self.next_poll_allowed - now)).await;
        }
        Ok(())
    }

    async fn sleep_until_ratelimit(&self) -> Result<(), Box<dyn std::error::Error>> {
        match self.ratelimit_remaining {
            Some(remaining) if remaining <= 0 => {}
            _ => {
                // If the number of remaining requests is either
                // unknown (None) or greater than 0, we do not need to
                // sleep.
                return Ok(());
            }
        };
        let reset = self
            .ratelimit_reset
            .expect("The above match statement proves this is a Some()");
        loop {
            let now = SystemTime::now().duration_since(UNIX_EPOCH)?.as_secs();
            if now >= reset {
                break;
            }
            sleep(Duration::from_secs(reset - now)).await;
        }
        Ok(())
    }
}

fn number_header(header: Option<&HeaderValue>) -> Result<Option<u64>, Box<dyn std::error::Error>> {
    Ok(header
        .map(|hdr| hdr.to_str())
        .transpose()?
        .map(|hdr| hdr.parse())
        .transpose()?)
}
